import os

INTERNAL_QUEUE_NAME = "_dbos_internal_queue"


class GlobalParams:
    app_version: str = os.environ.get("DBOS__APPVERSION", "")
    executor_id: str = os.environ.get("DBOS__VMID", "local")
